/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssMessage;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.ibv.simbaustelle.verwaltung.BssStau;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation.Daten.StauEintrag;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * berechnet die für eine Baustelle zu erwartetenden Staus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BaustellenSimulator {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Baustelle, für die die Berechnung erfolgen soll.
	 */
	private final BaustellenSchnittstelle baustelle;

	/**
	 * die Parameter für die Berechnung.<br>
	 * Das Objekt wird aus der übergeordneten Baustellenverwaltung übernommen.
	 */
	private final IBssParameterObserver parameter;

	/** die Liste der ermittelten Staus. */
	private final List<BssStau> stauListe = new ArrayList<>();

	/** der Name des Auftraggebers der Simulation. */
	private final String auftragGeber;

	/** eine Bemerkung zum Simulationsauftrag. */
	private final String bemerkung;

	/** optionale vorgegebener Mq vor der Baustelle. */
	private final MessQuerschnittAllgemein berechnungsMq;

	/**
	 * erzeugt eine Instanz der Klasse zur Berechnung von Staus in einer
	 * Baustelle.
	 *
	 * @param baustelle
	 *            die betroffenen Baustelle
	 * @param auftragGeber
	 *            der Name des Auftraggebers der Simulation
	 * @param bemerkung
	 *            eine Bemerkung zum Simulationsauftrag
	 * @param berechnungsMq
	 *            der optionale MQ für den die Berechnung ausgeführt werden soll
	 *            (als MQ vor der Baustelle)
	 * @param parameter
	 *            die Parameter für die Berechnung aus der Baustellenverwaltung
	 */
	public BaustellenSimulator(final BaustellenSchnittstelle baustelle, final String auftragGeber,
			final String bemerkung, final MessQuerschnittAllgemein berechnungsMq,
			final IBssParameterObserver parameter) {

		BaustellenSimulator.LOGGER.finer("Erzeuge BaustellenSimulator für Baustelle: " + baustelle.getName());

		this.baustelle = baustelle;
		this.parameter = parameter;
		this.auftragGeber = auftragGeber;
		this.bemerkung = bemerkung;
		this.berechnungsMq = berechnungsMq;

	}

	/**
	 * überträgt die Liste der berechneten Staus in einen
	 * Datenverteiler-Datensatz und stellt damit das Ergbnis zur
	 * Veröffentlichung zur Verfügung.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @param datenSatz
	 *            der Datensatz, in dem die Ergebnisse abgelegt werden sollen.
	 * @param manuell
	 *            Simulation manuell ausführen
	 */
	public final void ermittleDaten(final ClientDavInterface verbindung, final OdBaustellenSimulation.Daten datenSatz,
			final boolean manuell) {

		boolean simulationErfolgreich = false;
		String fehlerMeldung = null;

		if (baustelle.pruefeSimulierbarkeit(manuell)) {

			try {
				final BerechnungsZeitRaumSchnittstelle zeitRaum = new BerechnungsZeitRaum(System.currentTimeMillis(),
						baustelle, parameter, manuell);

				if (zeitRaum.isValid()) {
					stauListe.clear();
					if (zeitRaum.prognoseMoeglich()) {
						final ZufliessenderVerkehrSchnittstelle zufVerkehr = new ZufliessenderVerkehr(verbindung,
								baustelle, berechnungsMq, zeitRaum);
						stauListe.addAll(prognoseErstellen(zeitRaum, zufVerkehr));
					} else {
						BaustellenSimulator.LOGGER.info("Prognosestart- und -endzeit identisch");
						fehlerMeldung = "Prognosestart- und -endzeit identisch";
					}
					simulationErfolgreich = true;
					BaustellenSimulator.LOGGER.finer("Liste der Stauobjekte für " + baustelle.getName()
					+ " ermittelt - Anzahl: " + stauListe.size());
				} else {
					fehlerMeldung = MessageFormat.format(BssMessage.WarningUngueltigerSimulationsZeitraum.name(),
							baustelle, zeitRaum);
					LogTools.log(BaustellenSimulator.LOGGER, new BetriebsmeldungDaten(baustelle.getObjekt()),
							BssMessage.WarningUngueltigerSimulationsZeitraum, baustelle, zeitRaum);
				}
			} catch (final UndefiniertException e) {
				fehlerMeldung = MessageFormat.format(BssMessage.WarningFehlendeSimulationsDaten.name(), baustelle,
						e.getMessage());
				LogTools.log(BaustellenSimulator.LOGGER, new BetriebsmeldungDaten(baustelle.getObjekt()),
						BssMessage.WarningFehlendeSimulationsDaten, baustelle, e.getMessage());
			}
		}

		datenSatz.setSimulationErfolgreich(simulationErfolgreich);
		datenSatz.setAuftragGeber(auftragGeber);
		datenSatz.setBemerkung(bemerkung);
		datenSatz.setFehlerMeldung(fehlerMeldung);
		for (final BssStau stau : stauListe) {
			final StauEintrag eintrag = new OdBaustellenSimulation.Daten.StauEintrag();
			eintrag.setStartZeit(stau.getStartZeit());
			eintrag.setDauer(stau.getDauer());
			eintrag.setMaxLaenge(stau.getMaxLaenge());
			eintrag.setMaxLaengeZeit(stau.getMaxLaengeZeit());
			eintrag.setVerlustZeit(stau.getVerlustZeit());
			datenSatz.addSchritte(eintrag);
		}
	}

	/**
	 * ermittelt die Liste der Staus, die innerhalb des Prognosezeitraums zu
	 * erwarten sind.
	 *
	 * @param zeitRaum
	 *            der Zeitraum, für den die Staus prognostiziert werden sollen
	 * @param zufVerkehr
	 *            der zufließende Verkehr
	 * @return die Liste der ermittelten Staus
	 * @throws UndefiniertException
	 *             ein für die Prognose erforderlicher Wert konnte nicht
	 *             ermittelt werden, die Prognose wurde nicht ausgeführt.
	 */
	final List<BssStau> prognoseErstellen(final BerechnungsZeitRaumSchnittstelle zeitRaum,
			final ZufliessenderVerkehrSchnittstelle zufVerkehr) throws UndefiniertException {
		final List<BssStau> result = new ArrayList<>();

		final long stBeginn = parameter.getLaengeStauBeginn();
		final long stAufhebung = parameter.getLaengeStauAufhebung();
		long aktuelleLaenge = 0;
		long fzImStau = 0;
		BssStau aktuellerStau = null;

		final double schrittweite = ((double) parameter.getPrognoseSchrittweite() / Constants.MILLIS_PER_HOUR);

		BaustellenSimulator.LOGGER
		.finer(baustelle.getName() + ": Beginne Stauberechnung mit " + " Parameter 'LängeStaubeginn': "
				+ stBeginn + ", 'LängeStauAufhebung': " + stAufhebung + ", 'Schrittweite': " + schrittweite);

		for (final long zs : zeitRaum.getIntervalle()) {
			try {
				final double zufluss = zufVerkehr.getFzg(zs);
				final double restKapazitaet = baustelle.getDaten().getRestKapazitaet(zs);

				BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs) + ": Zufluss: "
						+ zufluss + ", Restkapazität: " + restKapazitaet);

				final double fzDiff = (zufluss - restKapazitaet) * schrittweite;

				final int fahrstreifenAnzahl = baustelle.anzahlFahrStreifen(aktuelleLaenge);
				final long laengePkw = parameter.getLaengePkw();

				final long lDiff = (long) ((fzDiff / fahrstreifenAnzahl) * laengePkw);
				aktuelleLaenge = Math.max(0, aktuelleLaenge + lDiff);
				fzImStau = Math.max(0, fzImStau + (long) fzDiff);
				final long verlustZeit = (long) (((double) fzImStau / baustelle.getDaten().getRestKapazitaet(zs))
						* Constants.MILLIS_PER_HOUR);

				BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs) + ": FzDiff: "
						+ fzDiff + ", FzImStau: " + fzImStau + ", Verlustzeit: " + verlustZeit);

				if (aktuellerStau == null) {
					if (aktuelleLaenge > stBeginn) {
						BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
						+ ": neuen Stau erzeugen: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau = new BssStau(zs, aktuelleLaenge, verlustZeit);
						result.add(aktuellerStau);
					} else {
						BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
						+ ": es wird kein Stau erzeugt, da ermittelte Staulänge (" + aktuelleLaenge
						+ "m) < Parameter 'LängeStaubeginn' (" + stBeginn + "m)");
					}
				} else {
					if (aktuelleLaenge < stAufhebung) {
						BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
						+ ": Stau beenden: Länge (" + aktuelleLaenge + "m) < Parameter 'LängeStauAufhebung' ("
						+ stAufhebung + "m)");
						aktuellerStau.beenden(zs);
						aktuellerStau = null;
					} else {
						BaustellenSimulator.LOGGER.finer(baustelle.getName() + ": " + getBerechnungZeitText(zs)
						+ ": Stau aktualisieren: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau.aktualisieren(zs, aktuelleLaenge, verlustZeit);
					}
				}
			} catch (final UndefiniertException ex) {
				if (!BaustellenSimulation.getInstance().isAllowGaps()) {
					throw ex;
				}
			}
		}

		return result;
	}

	/**
	 * Gibt den aktuellen Zeitpunkt der Berechnung als text zur&uuml;ck.
	 *
	 * @param zeit
	 *            Zeit in ms
	 * @return Zeitpunkt als Text
	 */
	private String getBerechnungZeitText(final long zeit) {
		return "Zeitpunkt: " + DateFormat.getDateTimeInstance().format(new Date(zeit));
	}
}
