/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import com.bitctrl.Constants;

import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation für den Berechnungszeitraum einer Baustellen-Stauprognose.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BerechnungsZeitRaum implements BerechnungsZeitRaumSchnittstelle {

	/**
	 * Logger für Debug-Ausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * der Startzeitpunkt des Zeitraums.<br>
	 * Der Startzeitpunkt ergibt sich aus dem Maximum des Tagesbeginns der
	 * Gültigkeit der Baustelle und dem Tagesbeginn des Prognosezeitpunktes.
	 */
	private final long startZeitPunkt;

	/**
	 * der Endzeitpunkt des Zeitraums.<br>
	 * Der Endzeitpunkt ergibt sich aus dem Minimum von Tagesende des
	 * Gültigkeitszeitraums der Baustelle und dem Tagesende des
	 * Prognosezeitraums (-horizonts).
	 */
	private long endZeitPunkt;

	/**
	 * die Schrittweite in Millisekunden.<br>
	 * Die Schrittweite wird als Parameter für die Baustellensimulation
	 * definiert.
	 */
	private final long schrittWeite;

	/**
	 * erzeugt eine Instanz des BerechnungsZeitraums auf Basis des übergebenen
	 * Zeitpunkts, den Baustelleninformationen und den allgemeinen
	 * Simulationsparametern.
	 *
	 * @param zeit
	 *            der Bezugszeitpunkt
	 * @param baustelle
	 *            die betroffene Baustelle
	 * @param parameter
	 *            die allgemeinen Simulationsparameter
	 * @param manuell
	 *            Berechnung wurde manuell beauftragt
	 * @throws UndefiniertException
	 *             der Berechnungszeitraum konnte nicht bestimmt werden, weil
	 *             nicht alle Parameter verfügbar waren
	 */
	public BerechnungsZeitRaum(final long zeit, final BaustellenSchnittstelle baustelle,
			final IBssParameterObserver parameter, final boolean manuell) throws UndefiniertException {

		if (manuell && (!BaustellenSimulation.getInstance().isZeitbereichAfoStrict())) {
			final Calendar cal = Calendar.getInstance();
			cal.setTimeInMillis(baustelle.getDaten().getStart());
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 0);
			startZeitPunkt = cal.getTimeInMillis();

			cal.setTimeInMillis(baustelle.getDaten().getEnde());
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 0);
			endZeitPunkt = cal.getTimeInMillis() + (24 * 3600 * 1000);

			final long manuellMsecs = BaustellenSimulation.getInstance().getMaxHorizontManuellMsecs();
			if ((manuellMsecs > 0) && ((endZeitPunkt - startZeitPunkt) > manuellMsecs)) {
				endZeitPunkt = startZeitPunkt + manuellMsecs;
			}

		} else {

			final Calendar hCal = new GregorianCalendar();
			hCal.setTimeInMillis(zeit);

			final Calendar bCal = new GregorianCalendar();
			bCal.setTimeInMillis(baustelle.getDaten().getStart());

			// /Anfangszeitpunkt ermitteln
			long horizontZeit = new GregorianCalendar(hCal.get(Calendar.YEAR), hCal.get(Calendar.MONTH),
					hCal.get(Calendar.DAY_OF_MONTH)).getTimeInMillis();
			BerechnungsZeitRaum.LOGGER
			.finer("Horizont-Startzeit: " + DateFormat.getDateTimeInstance().format(new Date(horizontZeit)));

			long bstZeit = new GregorianCalendar(bCal.get(Calendar.YEAR), bCal.get(Calendar.MONTH),
					bCal.get(Calendar.DAY_OF_MONTH)).getTimeInMillis();
			BerechnungsZeitRaum.LOGGER
			.finer("Baustellen-Startzeit: " + DateFormat.getDateTimeInstance().format(new Date(bstZeit)));

			startZeitPunkt = Math.max(horizontZeit, bstZeit);
			BerechnungsZeitRaum.LOGGER
			.finer("Prognosestart: " + DateFormat.getDateTimeInstance().format(new Date(startZeitPunkt)));

			// Endzeitpunkt ermitteln
			hCal.add(Calendar.SECOND, (int) (parameter.getHorizont() / Constants.MILLIS_PER_SECOND));
			horizontZeit = new GregorianCalendar(hCal.get(Calendar.YEAR), hCal.get(Calendar.MONTH),
					hCal.get(Calendar.DAY_OF_MONTH)).getTimeInMillis();
			horizontZeit += Constants.MILLIS_PER_DAY;

			bCal.setTimeInMillis(baustelle.getDaten().getEnde());
			bstZeit = new GregorianCalendar(bCal.get(Calendar.YEAR), bCal.get(Calendar.MONTH),
					bCal.get(Calendar.DAY_OF_MONTH)).getTimeInMillis();
			bstZeit += Constants.MILLIS_PER_DAY;

			endZeitPunkt = Math.max(startZeitPunkt, Math.min(horizontZeit, bstZeit));
			BerechnungsZeitRaum.LOGGER
			.finer("Prognoseende: " + DateFormat.getDateTimeInstance().format(new Date(endZeitPunkt)));
		}

		schrittWeite = parameter.getPrognoseSchrittweite();
	}

	@Override
	public final long getEndeZeitpunkt() {
		return endZeitPunkt;
	}

	@Override
	public final List<Long> getIntervalle() {
		final List<Long> result = new ArrayList<>();
		long wert = startZeitPunkt;

		if ((endZeitPunkt > startZeitPunkt) && (schrittWeite > 0)) {
			while (wert < endZeitPunkt) {
				result.add(wert);
				wert += schrittWeite;
			}
		}

		return result;
	}

	@Override
	public final long getStartZeitpunkt() {
		return startZeitPunkt;
	}

	@Override
	public boolean isValid() {
		return endZeitPunkt >= startZeitPunkt;
	}

	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer();
		buffer.append('(');
		buffer.append(DateFormat.getDateTimeInstance().format(new Date(startZeitPunkt)));
		buffer.append('-');
		buffer.append(DateFormat.getDateTimeInstance().format(new Date(endZeitPunkt)));
		buffer.append(", Schrittweite: ");
		buffer.append(schrittWeite);
		buffer.append(" ms)");
		return buffer.toString();
	}

	@Override
	public boolean prognoseMoeglich() {
		return endZeitPunkt > startZeitPunkt;
	}
}
