/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssKonstanten;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

/**
 * eine Hilfsobjekt zur Verarbeitung der Kommandos zur Ausführung einer
 * Simulation auf Nutzeranforerung.<br>
 * Die Klasse wertet die über die Attributgruppe
 * "atg.baustellenSimulationStarten" empfangenen Daten aus und löst eine
 * sofortige Berechnung der entsprechenden Baustelle aus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BaustellenSimulationsObserver implements ClientReceiverInterface {

	/**
	 * sammelt die Aufträge und führt diese im Hintergrund aus, um die
	 * Simulation von der restlichen Datenverteilerkommunikation zu entkoppeln.
	 */
	private static class BaustellenSimulationAuftragsVerwalter extends Thread {

		/**
		 * die Repräsentation eines auszuführenden Simulationsauftrags.
		 */
		private class Auftrag {

			/**
			 * die Baustelle, für die eine Simulation ausgeführt werden soll.
			 */
			private final BaustellenSchnittstelle baustelle;

			/** der Name des Auftraggebers für die Simulation. */
			private final String name;

			/** ein beschreibender Text für den Simulationsauftrag. */
			private final String bemerkung;

			/** der optional vorgegebene MQ vor der Baustelle. */
			private final MessQuerschnittAllgemein berechnungsMq;

			/**
			 * Konstruktor. Die Funktion legt einen Auftrag für die übergebene
			 * Baustelle mit den übergebenen Auftraggeberinformationen an.
			 *
			 * @param baustelle
			 *            die Baustelle
			 * @param name
			 *            der Name des Auftraggebers
			 * @param bemerkung
			 *            beschreibender Text
			 */
			Auftrag(final BaustellenSchnittstelle baustelle, final String name, final String bemerkung,
					final MessQuerschnittAllgemein berechnungsMq) {
				this.baustelle = baustelle;
				this.name = name;
				this.bemerkung = bemerkung;
				this.berechnungsMq = berechnungsMq;

			}
		}

		/**
		 * globale Instanz der Klasse.
		 */
		private static final BaustellenSimulationAuftragsVerwalter INSTANZ = new BaustellenSimulationAuftragsVerwalter();

		/**
		 * liefert die SINGLETON-Instanz der Klasse.
		 *
		 * @return die Instanz
		 */
		public static BaustellenSimulationAuftragsVerwalter getInstanz() {
			if (!BaustellenSimulationAuftragsVerwalter.INSTANZ.isAlive()) {
				BaustellenSimulationsObserver.LOGGER.finer("Auftragsverwalter nicht aktiv --- STARTEN");
				BaustellenSimulationAuftragsVerwalter.INSTANZ.start();
			} else {
				BaustellenSimulationsObserver.LOGGER.finest("Auftragsverwalter ist bereits aktiv");
			}
			return BaustellenSimulationAuftragsVerwalter.INSTANZ;
		}

		/**
		 * die Liste der auszuführenden Aufträge.
		 */
		private final List<Auftrag> auftragsListe = new ArrayList<>();

		/**
		 * der Status des Verwalters.
		 */
		private final boolean aktiv = true;

		/**
		 * Konstruktor.
		 */
		BaustellenSimulationAuftragsVerwalter() {
			super("BaustellenSimulationsAuftragsVerwalter");
		}

		/**
		 * fügt der Auftragsverwaltung einen neuen Auftrag hinzu.
		 *
		 * @param baustelle
		 *            die Baustelle die simuliert werden soll
		 * @param name
		 *            der Name des Auftraggebers
		 * @param bemerkung
		 *            ein beschreibender Text
		 */
		public void addAuftrag(final BaustellenSchnittstelle baustelle, final String name, final String bemerkung) {
			addAuftrag(baustelle, name, bemerkung, null);
		}

		/**
		 * fügt der Auftragsverwaltung einen neuen Auftrag hinzu.
		 *
		 * @param baustelle
		 *            die Baustelle die simuliert werden soll
		 * @param name
		 *            der Name des Auftraggebers
		 * @param bemerkung
		 *            ein beschreibender Text
		 * @param berechnungsMq
		 *            der MQ vor der Baustelle (optional)
		 */
		public void addAuftrag(final BaustellenSchnittstelle baustelle, final String name, final String bemerkung,
				final MessQuerschnittAllgemein berechnungsMq) {
			synchronized (auftragsListe) {
				auftragsListe.add(new Auftrag(baustelle, name, bemerkung, berechnungsMq));
			}
			synchronized (this) {
				this.notifyAll();
			}
		}

		@Override
		public void run() {
			while (aktiv) {
				Auftrag auftrag = null;
				synchronized (auftragsListe) {
					if (auftragsListe.size() > 0) {
						auftrag = auftragsListe.get(0);
						auftragsListe.remove(0);
					} else {
						BaustellenSimulationsObserver.LOGGER.finest("Auftragsliste ist leer");
					}
				}
				if (auftrag != null) {
					final String messageText;
					if (auftrag.berechnungsMq == null) {
						messageText = "Manuelle Simulation der Baustelle: " + auftrag.baustelle.getName()
						+ " angefordert von " + auftrag.name + ", Bemerkung: " + auftrag.bemerkung;
					} else {
						messageText = "Manuelle Simulation der Baustelle: " + auftrag.baustelle.getName()
						+ " angefordert von " + auftrag.name + ", Bemerkung: " + auftrag.bemerkung
						+ " Berechnungs-MQ: " + auftrag.berechnungsMq;
					}
					BaustellenSimulationsObserver.LOGGER.fine(messageText);
					MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.INFORMATION,
							messageText);

					auftrag.baustelle.simuliere(auftrag.name, auftrag.bemerkung, auftrag.berechnungsMq, true);
				} else {
					synchronized (this) {
						try {
							if (auftragsListe.size() <= 0) {
								this.wait(60000);
							}
						} catch (final InterruptedException e) {
							Debug.getLogger().fine(e.getLocalizedMessage());
							// Schleife wird wiederholt und eventuelle Aufträge
							// vorfristig ausgeführt
						}
					}
				}
			}
		}
	}

	/**
	 * Logger für Debugausgaben.
	 */
	static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Attributgruppe zum Empfang des Kommandos zum Start der Simulation.
	 */
	private static AttributeGroup startAtg;

	/**
	 * die Attributgruppe zum Empfang des Kommandos zum Start der Simulation mit
	 * vorgegebenem MQ.
	 */
	private static AttributeGroup startMitMqAtg;

	/**
	 * der erwartetet Aspekt.
	 */
	private static Aspect startAsp;

	/**
	 * die Baustelle, die gesteuert werden soll.
	 */
	private final Baustelle baustelle;

	/**
	 * erzeugt eine Instanz des Steuerungsobjekts.
	 *
	 * @param baustelle
	 *            die zu steuernde Baustelle
	 */
	public BaustellenSimulationsObserver(final Baustelle baustelle) {
		this.baustelle = baustelle;
		if (BaustellenSimulationsObserver.startAtg == null) {
			final DataModel model = baustelle.getVerbindung().getDataModel();
			BaustellenSimulationsObserver.startAtg = model
					.getAttributeGroup(BssKonstanten.PID_ATG_BAUSTELLEN_SIMULATION_STARTEN);
			BaustellenSimulationsObserver.startMitMqAtg = model
					.getAttributeGroup(BssKonstanten.PID_ATG_BAUSTELLEN_SIMULATION_STARTEN_MIT_MQ);
			BaustellenSimulationsObserver.startAsp = model.getAspect(BssKonstanten.PID_ASP_SENDEN);
		}

		baustelle.getVerbindung().subscribeReceiver(this, baustelle.getSystemObjekt(),
				new DataDescription(BaustellenSimulationsObserver.startAtg, BaustellenSimulationsObserver.startAsp),
				ReceiveOptions.normal(), ReceiverRole.drain());

		if (BaustellenSimulationsObserver.startMitMqAtg != null) {
			baustelle.getVerbindung()
			.subscribeReceiver(this, baustelle.getSystemObjekt(),
					new DataDescription(BaustellenSimulationsObserver.startMitMqAtg,
							BaustellenSimulationsObserver.startAsp),
					ReceiveOptions.normal(), ReceiverRole.drain());
		}
	}

	/**
	 * Die Funktion empfängt den Datensatz zur Ausführung einer Simulation und
	 * führt die Simulation der Baustelle aus.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public final void update(final ResultData[] results) {
		for (final ResultData result : results) {
			BaustellenSimulationsObserver.LOGGER.finer("Befehl zum manuellen Simulationsstart empfangen", result);

			final Data data = result.getData();
			if (data != null) {
				if (result.getDataDescription().getAttributeGroup().equals(BaustellenSimulationsObserver.startAtg)) {
					final String name = data.getTextValue(BssKonstanten.ATT_NAME_SIMSTART_NAME).getText();
					final String bemerkung = data.getTextValue(BssKonstanten.ATT_NAME_SIMSTART_BEMERKUNG).getText();
					BaustellenSimulationAuftragsVerwalter.getInstanz().addAuftrag(baustelle, name, bemerkung);
				} else if (result.getDataDescription().getAttributeGroup()
						.equals(BaustellenSimulationsObserver.startMitMqAtg)) {
					final String name = data.getTextValue(BssKonstanten.ATT_NAME_SIMSTART_NAME).getText();
					final String bemerkung = data.getTextValue(BssKonstanten.ATT_NAME_SIMSTART_BEMERKUNG).getText();
					final SystemObject mqObj = data.getReferenceValue(BssKonstanten.ATT_NAME_SIMSTART_BERECHNUNGS_MQ)
							.getSystemObject();

					MessQuerschnittAllgemein berechnungsMq = null;
					if (mqObj != null) {
						berechnungsMq = (MessQuerschnittAllgemein) ObjektFactory.getInstanz().getModellobjekt(mqObj);
					}
					BaustellenSimulationAuftragsVerwalter.getInstanz().addAuftrag(baustelle, name, bemerkung,
							berechnungsMq);
				}
			}
		}
	}

	/** entfernt die Datenanmeldungen des Observers. */
	public void disconnect() {
		baustelle.getVerbindung().unsubscribeReceiver(this, baustelle.getSystemObjekt(),
				new DataDescription(BaustellenSimulationsObserver.startAtg, BaustellenSimulationsObserver.startAsp));
		if (BaustellenSimulationsObserver.startMitMqAtg != null) {
			baustelle.getVerbindung().unsubscribeReceiver(this, baustelle.getSystemObjekt(), new DataDescription(
					BaustellenSimulationsObserver.startMitMqAtg, BaustellenSimulationsObserver.startAsp));
		}
	}
}
